#include "demo1.h"
#include "Shader.h"

using namespace std;

//window resolution (default 640x480)
float windowX = 640.0f;
float windowY = 480.0f;

//The mesh global variable
TriangleMesh trig;

//The shader used to render the scene
Shader shader;

//Camera matrices
glm::mat4 projectionMatrix; //The view projection matrix (orthographic or perspective)
glm::mat4 viewMatrix; //The transformation matrix for the virtual camera
glm::mat4 modelMatrix; //The trasnformation matrix for the teapot model

//The vertex buffer object used for holding vertex information (from teapot mesh) for the shader
//You should have one of these for each component for rendering e.g. positions, normals, UV (texture) coordinates
unsigned int vbo;

void cleanup()
{
	//do any cleanup here
}

//This function loads an obj format file
//This is a utility and should not have to be modified for teapot.obj (the assignment).
//Note that this function does not load the normals for the mesh.  
//These are not provided in the teapot.obj file, you must compute these yourself
void TriangleMesh::LoadFile(char * filename)
{
	FILE *file = fopen(filename, "r");
	if(file  == NULL){
		std::cout << "Can't open file " << filename << std::endl;
		return;
	}

	std::vector<unsigned int> vertexIndices, uvIndices;
	std::vector<glm::vec3> tempVertices;
	std::vector<glm::vec2> tempUVs;

	_max.x =-10000; _max.y =-10000; _max.z =-10000;
	_min.x =10000; _min.y =10000; _min.z =10000;
	glm::vec3 averageVertex(0.0f);	

	//Read in .obj
	while(1){
		char lineHeader[128];
		//read the first word of the line
		int res = fscanf(file, "%s", lineHeader);
		if(res == EOF){
			break;
		}
		if(strcmp(lineHeader, "v") == 0){
			glm::vec3 vertex;
			fscanf(file, "%f %f %f\n", &vertex.x, &vertex.y, &vertex.z);
			tempVertices.push_back(vertex);

			averageVertex += vertex;

			if (vertex.x > _max.x) _max.x = vertex.x;
			if (vertex.y > _max.y) _max.y = vertex.y;
			if (vertex.z > _max.z) _max.z = vertex.z;

			if (vertex.x < _min.x) _min.x = vertex.x;
			if (vertex.y < _min.y) _min.y = vertex.y;
			if (vertex.z < _min.z) _min.z = vertex.z;

		}
		else if(strcmp(lineHeader, "vt" ) == 0){
			glm::vec2 uv;
			fscanf(file, "%f %f\n", &uv.x, &uv.y);
			tempUVs.push_back(uv);
		}
		else if(strcmp(lineHeader, "f" ) == 0){
			std::string vertex1, vertex2, vertex3;
			unsigned int vertexIndex[3], uvIndex[3];
			int matches = fscanf(file, "%d/%d %d/%d %d/%d\n",&vertexIndex[0], &uvIndex[0], &vertexIndex[1], &uvIndex[1], &vertexIndex[2], &uvIndex[2] );
			if(matches != 6){
				std::cout << "Can't be read by simple parser!" << std::endl;
				return;
			}
			Triangle trig(vertexIndex[0],vertexIndex[1],vertexIndex[2]);
			_triangles.push_back(trig);

			vertexIndices.push_back(vertexIndex[0]);
			vertexIndices.push_back(vertexIndex[1]);
			vertexIndices.push_back(vertexIndex[2]);
			uvIndices.push_back(uvIndex[0]);
			uvIndices.push_back(uvIndex[1]);
			uvIndices.push_back(uvIndex[2]);
		}
	}		

	//Arrange in triangle order to fit with OpenGL's glDrawArrays(...) function. 
	//i.e. _vertices[0],_vertices[1], and _vertices[2] are the first triangle, 
	//_vertices[3],_vertices[4], and _vertices[5] are the second etc.
	//Note this invalidates the Triangle structures in _triangles.  The Triangle class 
	//is used to help with creating an array of indexes for glDrawElements(...); if you want to use 
	//this instead of reordering the vertices
	_vertices.resize(vertexIndices.size());
	_uvs.resize(uvIndices.size());
	for(unsigned int i = 0; i < vertexIndices.size(); ++i){
		_vertices[i] = tempVertices[vertexIndices[i]-1];
		_uvs[i] = tempUVs[uvIndices[i]-1];		
	}

	float range; 
	if (_max.x-_min.x > _max.y-_min.y){
		range = _max.x-_min.x;
	}
	else{
		range = _max.y-_min.y;
	}

	averageVertex /= _vertices.size();

	for (int i = 0; i < _vertices.size(); i++) 
	{
		_vertices[i] = (_vertices[i]-averageVertex)/range*400.0f;  
	}

	cout << "Number of triangles: " << _triangles.size() << ", number of vertices: " << _vertices.size() << endl;
};

/*--- Display Function ---*/
//The main display function.
//This allows you to draw pixels onto the display by using GL_POINTS.
//Drawn every time an update is required.
//Students: This is the main file you'll need to modify or replace.
//The idea with this example function is the following:
//1)Clear the screen so we can draw a new frame
//2)Calculate the vertex points for each triangle and draw them (vertices only)
//3)Flush the pipeline so that the instructions we gave are performed.
void DemoDisplay()
{
	glClear(GL_COLOR_BUFFER_BIT); // Clear OpenGL Window

	glColor3f(1,1,1);  // The colour we will be drawing is white (red = 1, green = 1, blue = 1).

	//Tell OpenGL to use the shader we have created
	shader.Bind();

	//Find the location of our uniform variables in the current shader program
	int projectMatrixLocation = glGetUniformLocation(shader.ID(), "projectionMatrix");
	int viewMatrixLocation = glGetUniformLocation(shader.ID(), "viewMatrix");
	int modelMatrixLocation = glGetUniformLocation(shader.ID(), "modelMatrix");		
	//Pass the current values for our variables to the shader program
	glUniformMatrix4fv(projectMatrixLocation, 1, GL_FALSE, &projectionMatrix[0][0]);
	glUniformMatrix4fv(viewMatrixLocation, 1, GL_FALSE, &viewMatrix[0][0]);
	glUniformMatrix4fv(modelMatrixLocation, 1, GL_FALSE, &modelMatrix[0][0]);	

	//Find the location for our vertex position variable
	const char * attribute_name = "in_position";
	int positionLocation = glGetAttribLocation(shader.ID(), attribute_name);
	if (positionLocation == -1) {
		std::cout << "Could not bind attribute " << attribute_name << std::endl;
		return;
	}
	//Tell OpenGL we will be using vertex position variable in the shader
	glEnableVertexAttribArray(positionLocation);
	//Bind our vertex position buffer
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	//Define how to use our vertex buffer object.  This applies to whatever VBO is currently bound to GL_ARRAY_BUFFER
	glVertexAttribPointer(
		positionLocation, // attribute (location of the in_position variable in our shader program)
		3,                 // number of elements per vertex, here (x,y,z)
		GL_FLOAT,          // the type of each element
		GL_FALSE,          // take our values as-is
		0,                 // no extra data between each position
		0                  // offset of first element
	);	

	//Do the actual rendering of the primitives using all active attribute arrays
	glDrawArrays(GL_TRIANGLES,0,trig.VertexCount());

	//Disable usage of the array
	glDisableVertexAttribArray(positionLocation);
	//Disable the shader program
	shader.Unbind();

	glFlush();// Output everything (write to the screen)
}


//This function is called when a (normal) key is pressed
//x and y give the mouse coordinates when a keyboard key is pressed
void DemoKeyboardHandler(unsigned char key, int x, int y)
{
	if(key == 'm')
	{
	cout << "Mouse location: " << x << " " << y << endl;
	}

	cout << "Key pressed: " << key << endl;

}

void SetupVBO()
{
	//Create a new Vertex Buffer Object
	glGenBuffers(1,&vbo);
	//Bind and stream data to the VBO 
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, //the target buffer object, this is why we bind our VBO here
		sizeof(glm::vec3)*trig.VertexCount(), //Size in bytes for the data
		&trig.Vertices()[0],//pointer to the array of data
		GL_STATIC_DRAW);//How to use the data (incorrect flag here could significantly harm performance, please read the OpenGL documentation).

	std::cout << "VBO generated!" << std::endl;	
}


//Program entry point.
//argc is a count of the number of arguments (including the filename of the program).
//argv is a pointer to each c-style string.
int main(int argc, char **argv)
{

	atexit(cleanup);
    cout << "Computer Graphics Assignment 1 Demo Program" << endl;

	if (argc >  1)  
	{
		trig.LoadFile(argv[1]);
	}
	else 
	{
        cerr << "Usage:" << endl;
		cerr << argv[0] << " <filename> " << endl;
		exit(1);
	}

	//initialise OpenGL
	glutInit(&argc, argv);
	//Define the window size with the size specifed at the top of this file
	glutInitWindowSize(windowX, windowY);	

	//Create the window for drawing with the title "SimpleExample"
	glutCreateWindow("CG-CW1");	

	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);	

	//Set the function demoDisplay (defined above) as the function that
	//is called when the window must display
	glutDisplayFunc(DemoDisplay);// Callback function
    //similarly for keyboard input
    glutKeyboardFunc(DemoKeyboardHandler);

	//Initialise the OpenGL Extension Wrangler library.  This lets us use Vertex Buffer Objects and appropriate extensions
	GLenum err = glewInit();
	if (err != GLEW_OK){
		std::cout << "Error!" << std::endl;
		exit(1); // or handle the error in a nicer way
	}
	if (!GLEW_VERSION_2_1) { // check that the machine supports the 2.1 API.
		std::cout << "Error 2.1!" << std::endl;
		exit(1); // or handle the error in a nicer way
	}

	//Create our shader
	shader.Init("shaders/exampleShader.vert","shaders/exampleShader.frag");	

	//Prepare data for OpenGL
	SetupVBO();

	//Set up our camera and object transformation matrices
	projectionMatrix = glm::ortho(-windowX*0.5f, windowX*0.5f, -windowY*0.5f,  windowY*0.5f, -1.0f, 400.0f);
	viewMatrix = glm::translate(glm::mat4(1.0f),glm::vec3(-20.0f,-50.0f,-20.0f));
	modelMatrix = glm::translate(glm::mat4(1.0f),glm::vec3(0.0f));

	//Run the GLUT internal loop
	glutMainLoop();// Display everything and wait
}
